/* Juliette.java by Wolf Jung (C) 2010.  www.mndynamics.com
   Defines the Applet class Juliette and the class Stringdialog.

   These classes are part of Juliette 1.0, which is free software; you can
   redistribute and / or modify them under the terms of the GNU General Public
   License as published by the Free Software Foundation, see www.gnu.org;
   either version 3, or (at your option) any later version. In short: there is
   no warranty of any kind; you must redistribute the source code as well.
*/
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.applet.*;

/* The applet Juliette draws two BufferedImages. These are computed in
   off-screen threads of the class Mnplane, which contains additional
   mathematical functions. Mouse events are dispatched manually. The
   variables Preper and Per always refer to the Misiurewicz point as*
   of pPlane, which is changed with x. Finding a center with x sets
   locper, which is used for local similarity p, but is reset to 0 by
   many commands. Then p will be using the same (pre-)period as t.

   * The appletviewer does not close when the applet is finished.
   * In Konqueror, the dialog box is badly positioned and sized.
     It does not give keyboard focus back to the applet. The
     text is not highlighted. There is a rare crash.
   * In Explorer, the images are distorted by scrolling.
   * The Stringdialog cannot be closed with ESC, only with Alt+F4. Now
     Enter does work, but not for the message box without TextField.
*/
public class Juliette extends Applet
implements MouseListener, KeyListener, ActionListener
{  static final long serialVersionUID = 0;
   protected int imax, kmax, offset, frame, locper, Preper, Per;
   protected transient BufferedImage pBuffer, dBuffer;
   protected transient Mnplane plane, pPlane, dPlane;
   protected PopupMenu popup;
   protected Menu levelMenu;
   protected MenuItem zoomin, zoomout, restart, nmaxi, exray,
      map, coords, find, renorm, locsim, assym,
      levela, level0, level1, level2, level3, levelm;

public Mnplane pPlaneExposed() //for javaScript
{ return pPlane; }

public Mnplane dPlaneExposed() //for javaScript
{ return dPlane; }

public void setParameter(double a, double b) //for javaScript
{ double[] c = {a, b};	pPlane.setPoint(c); drawRight(); }

public String getAppletInfo()
{  String s = "Juliette 1.0 by Wolf Jung (C), July 2010. www.mndynamics.com";
   return s;
}

public String[][] getParameterInfo()
{  String s[][] = {
      {"realPara", "double", "real part of the parameter c"},
      {"imagPara", "double", "imaginary part of parameter c"}
   }; return s;
}

public void init()
{  super.init();
   offset = getSize().width; imax = (offset - 30)/4; kmax = getSize().height/2;
   if (imax <= kmax + 5 && kmax <= imax + 5) imax = kmax;
   frame = kmax % 4; kmax -= frame; imax -= frame;
   if (imax < 4 || kmax < 4) System.exit(1);
   offset -= (2*imax + frame);

   pBuffer = new BufferedImage(2*imax, 2*kmax, BufferedImage.TYPE_3BYTE_BGR );
   pPlane = new Mnplane(imax, kmax, pBuffer, this);
   dBuffer = new BufferedImage(2*imax, 2*kmax, BufferedImage.TYPE_3BYTE_BGR );
   dPlane = new Mnplane(imax, kmax, dBuffer, this);

   double t[] = {-0.101096363845622, 0.956286510809142, //a
      1.328330115300301, 2.086595344738708, //abs rho, arg rho
      -0.035, 0.03}; //K
   pPlane.setScale(7, 1, t); Preper = 3; Per = 1;
   t[0] = 0.0; t[1] = 0.0; t[2] = 2.0; t[3] = 0.0; String s;
   s = getParameter("realPara"); if (s != null)
      try { t[0] = Double.parseDouble(s); }
         catch (NumberFormatException e) { t[0] = 0; }
   s = getParameter("imagPara"); if (s != null)
      try { t[1] = Double.parseDouble(s); }
         catch (NumberFormatException e) { t[1] = 0; }
   pPlane.setPoint(t); dPlane.setPoint(t); dPlane.setParameter(t);
   locper = pPlane.period(t); if (locper > 1024) locper = 0;
   if (locper > 0 && locper <= 3) { t[0] = 0; t[1] = 0; }
   dPlane.setPlane(t); dPlane.start();
   if (t[0] != 0 || t[1] != 0) { pPlane.setNmax(1000); dPlane.setNmax(1000); }
   if (locper > 0 && locper <= 3) { locper = 0; t[0] = -0.5; t[2] = 1.6; }
   pPlane.setPlane(t); pPlane.start();

   popup = new PopupMenu();

   zoomin = new MenuItem("z  Zoom in");
   zoomin.setActionCommand("zoomin"); zoomin.addActionListener(this);
   popup.add(zoomin);

   zoomout = new MenuItem("v  Zoom out");
   zoomout.setActionCommand("zoomout"); zoomout.addActionListener(this);
   popup.add(zoomout);

   restart= new MenuItem("h  Restart");
   restart.setActionCommand("restart"); restart.addActionListener(this);
   popup.add(restart);

   nmaxi = new MenuItem("n  Iterations ...");
   nmaxi.setActionCommand("nmaxi"); nmaxi.addActionListener(this);
   popup.add(nmaxi);

   popup.addSeparator();

   exray = new MenuItem("e  External ray ...");
   exray.setActionCommand("exray"); exray.addActionListener(this);
   popup.add(exray);

   map = new MenuItem("f  Map f_c(z)");
   map.setActionCommand("map"); map.addActionListener(this);
   popup.add(map);

   popup.addSeparator();

   coords = new MenuItem("i  Coordinates ...");
   coords.setActionCommand("coords"); coords.addActionListener(this);
   popup.add(coords);

   find = new MenuItem("x  Find point ...");
   find.setActionCommand("find"); find.addActionListener(this);
   popup.add(find);

   popup.addSeparator();

   renorm = new MenuItem("r  Renormalize ...");
   renorm.setActionCommand("renorm"); renorm.addActionListener(this);
   popup.add(renorm);

   locsim = new MenuItem("p  Local similarity");
   locsim.setActionCommand("locsim"); locsim.addActionListener(this);
   popup.add(locsim);

   assym = new MenuItem("t  Asymptotic scaling");
   assym.setActionCommand("assym"); assym.addActionListener(this);
   popup.add(assym);

   levelMenu = new Menu("Levels of asymptotic scaling");

   levela = new MenuItem("a  Self-similarity around a");
   levela.setActionCommand("levela"); levela.addActionListener(this);
   levelMenu.add(levela);

   level0 = new MenuItem("0  Scaling around c_n , \u03b3 = 1");
   level0.setActionCommand("level0"); level0.addActionListener(this);
   levelMenu.add(level0);

   level1 = new MenuItem("1  Scaling around c_n , \u03b3 = 3/2");
   level1.setActionCommand("level1"); level1.addActionListener(this);
   levelMenu.add(level1);

   level2 = new MenuItem("2  Scaling around c_n , \u03b3 = 7/4");
   level2.setActionCommand("level2"); level2.addActionListener(this);
   levelMenu.add(level2);

   level3 = new MenuItem("3  Scaling around c_n , \u03b3 = 15/8");
   level3.setActionCommand("level3"); level3.addActionListener(this);
   levelMenu.add(level3);

   levelm = new MenuItem("m  Little M-sets at c_n , \u03b3 = 2");
   levelm.setActionCommand("levelm"); levelm.addActionListener(this);
   levelMenu.add(levelm);

   popup.add(levelMenu);

   add(popup);
   enableEvents(AWTEvent.MOUSE_EVENT_MASK);
   addMouseListener(this);
   addKeyListener(this);
}

public void start()
{ plane = pPlane; pPlane.draw(1.0, 0, 0); dPlane.draw(1.0, 0, 0); }

public void stop()
{ pPlane.Stop(); dPlane.Stop(); }

public void destroy()
{  pPlane.Destroy(); pPlane = null; dPlane.Destroy(); dPlane = null;
   System.exit(0);
}

public void update(Graphics g)
{ paint(g); }

public void paint(Graphics g) //repaint() calls update() calls paint()
{  g.setColor(Color.gray);
   g.fillRect(2*imax+2*frame, 0, offset-2*imax-3*frame, 2*kmax+2*frame);
   if (frame > 0)
   {  if (plane == pPlane) g.setColor(Color.lightGray);
      else g.setColor(Color.darkGray);
      g.fillRect(0, 0, frame, 2*kmax+2*frame);
      g.fillRect(frame, 0, 2*imax+frame, frame);
      g.fillRect(offset, 2*kmax+frame, 2*imax+frame, frame);
      g.fillRect(offset+2*imax, frame, frame, 2*kmax+frame);
      if (plane == pPlane) g.setColor(Color.darkGray);
      else g.setColor(Color.lightGray);
      g.fillRect(frame, 2*kmax+frame, 2*imax+frame, frame);
      g.fillRect(2*imax+frame, frame, frame, 2*kmax);
      g.fillRect(offset-frame, 0, frame, 2*kmax+2*frame);
      g.fillRect(offset, 0, 2*imax+frame, frame);
   }
   synchronized (pBuffer) { g.drawImage(pBuffer, frame, frame, this); }
   synchronized (dBuffer) { g.drawImage(dBuffer, offset, frame, this); }
   g.setColor(Color.yellow); int[] ik = new int[2];
   pPlane.getPos(ik); g.setClip(frame, frame, 2*imax, 2*kmax);
   g.drawLine(frame + ik[0] - 4, frame + ik[1],
      frame + ik[0] + 4, frame + ik[1]);
   g.drawLine(frame + ik[0], frame + ik[1] - 4,
      frame + ik[0], frame + ik[1] + 4);
   dPlane.getPos(ik); g.setClip(offset, frame, 2*imax, 2*kmax);
   g.drawLine(offset + ik[0] - 4, frame + ik[1],
      offset + ik[0] + 4, frame + ik[1]);
   g.drawLine(offset + ik[0], frame + ik[1] - 4,
      offset + ik[0], frame + ik[1] + 4);
}

public void drawRight()
{  double[] t = new double[4];
   pPlane.getPoint(t); dPlane.setPoint(t); dPlane.setParameter(t);
   t[0] = 0.0; t[1] = 0.0; t[2] = 2.0; t[3] = 0.0; dPlane.setPlane(t);
   dPlane.setNmax(100); dPlane.draw(1.0, 0, 0);
}

public void processMouseEvent(MouseEvent e)
{  requestFocusInWindow();
   if (e.isPopupTrigger())
   {  if (e.getX() <= imax + offset/2) plane = pPlane; else plane = dPlane;
      popup.show(this, e.getX(), e.getY()); repaint();
   }
   else super.processMouseEvent(e);
}

public void mouseEntered(MouseEvent e) {}
public void mouseExited(MouseEvent e) {}
public void mouseClicked(MouseEvent e) {}
public void mousePressed(MouseEvent e) {}

public void mouseReleased(MouseEvent e)
{  int[] ik = new int[2]; ik[0] = e.getX(); ik[1] = e.getY();
   if (ik[0] < 2*imax + 2*frame)
   {  if (plane == dPlane) { plane = pPlane; repaint(); return; }
      ik[0] -= frame; ik[1] -= frame;
      if (ik[0] < 0 || ik[0] >= 2*imax || ik[1] < 0 || ik[1] >= 2*kmax) return;
      pPlane.setPos(ik); drawRight();
   }
   if (ik[0] >= offset - frame)
   {  if (plane == pPlane) { plane = dPlane; repaint(); return; }
      ik[0] -= offset; ik[1] -= frame;
      if (ik[0] < 0 || ik[0] >= 2*imax || ik[1] < 0 || ik[1] >= 2*kmax) return;
      dPlane.setPos(ik); repaint();
   }
}

public void keyPressed(KeyEvent e)
{  showStatus("");
   int key = e.getKeyCode();
   boolean noArrow = true;
   if (key == KeyEvent.VK_LEFT)  { plane.move(-1, 0); noArrow = false; }
   if (key == KeyEvent.VK_RIGHT) { plane.move(1,0); noArrow = false; }
   if (key == KeyEvent.VK_UP)    { plane.move(0, 1); noArrow = false; }
   if (key == KeyEvent.VK_DOWN)  { plane.move(0, -1); noArrow = false; }
   if (noArrow) return;
   if (plane == pPlane) drawRight(); else repaint();
}

public void keyReleased(KeyEvent e) {}

public void keyTyped(KeyEvent e)
{  char c = e.getKeyChar(); String s = "";
   if (c == 'z') s = "zoomin";
   if (c == 'v') s = "zoomout";
   if (c == 'h') s = "restart";
   if (c == 'n') s = "nmaxi";
   if (c == 'e') s = "exray";
   if (c == 'f') { s = "map"; plane = dPlane; }
   if (c == 'i') s = "coords";
   if (c == 'x') s = "find";
   if (c == 'r') s = "renorm";
   if (c == 'p') { s = "locsim"; plane = pPlane; }
   if (c == 't') { s = "assym"; plane = pPlane; }
   if (c == 'a') { s = "levela"; plane = pPlane; }
   if (c == '0') { s = "level0"; plane = pPlane; }
   if (c == '1') { s = "level1"; plane = pPlane; }
   if (c == '2') { s = "level2"; plane = pPlane; }
   if (c == '3') { s = "level3"; plane = pPlane; }
   if (c == 'm') { s = "levelm"; plane = pPlane; }
   performAction(s);
}

public void actionPerformed(ActionEvent e)
{ performAction(e.getActionCommand()); }

public void performAction(String s)
{  showStatus("");

   if (s.equals("zoomin")) plane.draw(0.64, 0, 0); //or 0.78125

   if (s.equals("zoomout")) plane.draw(1.5625, 0, 0); //or 1.28

   if (s.equals("restart"))
   {  if (plane == dPlane) { drawRight(); return; }
      double[] t = {-0.5, 0, 1.6, 0}; pPlane.setPlane(t);
      pPlane.setNmax(100); pPlane.draw(1.0, 0, 0); locper = 0;
   }

   if (s.equals("nmaxi"))
   {  int n = plane.getNmax(); long[] l = {(long) n};
      inputDialog(l, null, null,
         "Adjust the maximal number of iterations, 1...32000:");
      if (l[0] > 0 && l[0] <= 32000)
      plane.setNmax((int) l[0]); plane.draw(1.0, 0, 0);
   }

   if (s.equals("exray"))
   {  long[] l = {0L, 1L}; char[] sep = {'/'};
      inputDialog(l, null, sep, "Enter a rational angle:");
      if (l[0] >= 0 && l[0] < l[1]) plane.newtonRay(l[0], l[1]);
   }

   if (s.equals("map")) { dPlane.iterate(); repaint(); }

   if (s.equals("coords"))
   {  double[] z = new double[2]; plane.getPoint(z); char[] sep = {' '};
      inputDialog(null, z, sep, "Enter new coordinates:");
      if (z[0] < 1e10 && z[1] < 1e10) plane.setPoint(z);
      if (plane == pPlane) drawRight(); else repaint();
   }

   if (s.equals("find"))
   {  double[] t = new double[6]; pPlane.getPoint(t); t[4] = t[0]; t[5] = t[1];
      int q = pPlane.period(t); if (q > 1024) q = 0;
      int pp = 0, p = q;
      if (p == 0 || p > 1024) p = locper; if (p == 0) { pp = Preper; p = Per;}
      long[] l = {(long) pp, (long) p}; char[] sep = {','}; String text;
      if (plane == pPlane) text =
           "Enter the \"period\" of a center, or the\n"
         + "\"preperiod,period\" of a Misiurewicz point, 1...1024:\n"
         + "These values will be used by p (local similarity)\n"
         + "and by t (asymptotic scaling) as well.";
      else text =
           "Enter the \"period\" of a periodic point, or the\n"
         + "\"preperiod,period\" of a preperiodic point, 1...1024:";
      inputDialog(l, null, sep, text);
      pp = (int) l[0]; p = (int) l[1]; if (pp < 0 || p <= 0) return;
      plane.getPoint(t);
      if (plane.find(pp, p, t) != 0)
      { showStatus("Failed finding " + pp + ',' + p); return; }
      plane.setPoint(t); if (plane == dPlane) { repaint(); return; }
      drawRight(); if (pp == 0) { locper = p; return; }
      double[] rho = {t[0], t[1]}; if (pPlane.find(pp, -p, rho) != 0) return;
      t[2] = Math.sqrt(rho[0]*rho[0] + rho[1]*rho[1]);
      if (t[2] <= 1.0 || t[2] > 1.0e3) return;
      t[3] = Math.atan2(rho[1], rho[0]); t[4] -= t[0]; t[5] -= t[1];
      locper = 0; Preper = pp; Per = p; pPlane.setScale(q, p, t);
      pPlane.similarity(pp, -p, t); showStatus("\u03bb is  "+t[0]+"  "+t[1]);
   }

   if (s.equals("renorm"))
   {  double[] t = new double[4]; pPlane.getPoint(t);
      int pp = 0, p = pPlane.period(t); if (p == 0 || p > 1024) p = locper;
      if (p == 0) p = pPlane.scale(-1.0);
      long[] l = {(long) pp, (long) p}; char[] sep = {','};
      inputDialog(l, null, sep,
           "Enter the \"preperiod,period\" of renormalization:\n"
         + "(The preperiod is needed for embedded Julia sets.)");
      pp = (int) l[0]; p = (int) l[1]; if (pp < 0 || p <= 0) return;
      if (pPlane.similarity(0, -p, t) != 0) t[0] = 0.25;
      else
      {  t[1] = t[2]*t[2] + t[3]*t[3]; t[0] = 0.25;
         if (t[1] > 1.0) t[0] = 10.0/t[1];
      }
      plane.draw(t[0], pp, p);
   }

   if (s.equals("locsim"))
   {  double[] t = new double[4];
      int pp = 0, p = locper; if (p <= 0) p = pPlane.scale(-1.0);
      if (p <= 0) { pp = Preper; p = Per; }
      int n = pPlane.similarity(pp, p, t);
      if (n != 0) { showStatus("Error" + n); return; }
      n = pPlane.getNmax(); dPlane.setNmax(n);
      dPlane.setPlane(t); pPlane.getPoint(t); dPlane.setPoint(t);
      dPlane.draw(1.0, 0,0);
      showStatus("Dynamics rescaled with " + pp + ',' + p);
   }

   if (s.equals("assym"))
   {  pPlane.scale(0); pPlane.draw(1.0, 0, 0); locper = 0; drawRight();
      showStatus("Parameters rescaled n \u2192 n + 1");
   }

   if (s.equals("levela"))
   {  pPlane.scale(3.0); pPlane.draw(1.0, 0, 0); locper = 0; drawRight();
      showStatus("Scaling around a");
   }

   if (s.equals("level0"))
   {  pPlane.scale(1.0); pPlane.draw(1.0, 0, 0); locper = 0; drawRight();
      showStatus("Scaling with \u03b3 = 1");
   }

   if (s.equals("level1"))
   {  pPlane.scale(1.5); pPlane.draw(1.0, 0, 0); locper = 0; drawRight();
      showStatus("Scaling with \u03b3 = 3/2");
   }

   if (s.equals("level2"))
   {  pPlane.scale(1.75); pPlane.draw(1.0, 0, 0); locper = 0; drawRight();
      showStatus("Scaling with \u03b3 = 7/4");
   }

   if (s.equals("level3"))
   {  pPlane.scale(1.875); pPlane.draw(1.0, 0, 0); locper = 0; drawRight();
      showStatus("Scaling with \u03b3 = 15/8");
   }

   if (s.equals("levelm"))
   {  pPlane.scale(2.0); pPlane.draw(1.0, 0, 0); locper = 0; drawRight();
      showStatus("Scaling with \u03b3 = 2");
   }
} //performAction

//The inputDialog assumes that l or z has two elements when sep != null.
//It provides a message box when l and z are null.
public void inputDialog(long[] l, double[] z, char[] sep, String text)
{  String[] s = {""}; Point p = getLocationOnScreen();
   if (l != null)
   { s[0] = "" + l[0]; if (sep != null) s[0] += ("" + sep[0] + l[1]); }
   if (z != null)
   { s[0] = "" + z[0]; if (sep != null) s[0] += ("" + sep[0] + z[1]); }
   Stringdialog d = new Stringdialog(null, text, s, p.x + 30, p.y + 20);
   d.setVisible(true); requestFocusInWindow();
   int i = -1; if (sep != null) i = s[0].indexOf(sep[0]);
   if (l != null)
   {  if (i <= 0)
      {  try { l[0] = Long.parseLong(s[0]); }
            catch (NumberFormatException e) { l[0] = -1; }
         if (sep != null) { l[1] = l[0]; l[0] = 0; }
         return;
      }
      try { l[0] = Long.parseLong(s[0].substring(0, i)); }
         catch (NumberFormatException e) { l[0] = -1; }
      try { l[1] = Long.parseLong(s[0].substring(i + 1)); }
         catch (NumberFormatException e) { l[1] = -1; }
   }
   if (z != null)
   {  s[0].replace(',', '.');
      if (i <= 0)
      {  try { z[0] = Double.parseDouble(s[0]); }
            catch (NumberFormatException e) { z[0] = 1.0e100; }
         if (sep != null) z[1] = 1.0e100;
         return;
      }
      try { z[0] = Double.parseDouble(s[0].substring(0, i)); }
         catch (NumberFormatException e) { z[0] = 1.0e100; }
      try { z[1] = Double.parseDouble(s[0].substring(i + 1)); }
         catch (NumberFormatException e) { z[1] = 1.0e100; }
   }
} //inputDialog

} //Juliette


//Stringdialog shows the text, requests a String in s[0], has two buttons.
class Stringdialog extends Dialog implements ActionListener, WindowListener
{  static final long serialVersionUID = 0;
   protected transient TextArea ta;
   protected transient TextField tf;
   protected transient String[] s;
   protected transient Button b1, b2;

public Stringdialog(Frame parent, String text, String[] S, int x, int y)
{  super(parent, "Juliette", true); setBounds(x, y, 350, 250); s = S;
   BorderLayout layout = new BorderLayout(); setLayout(layout);
   ta = new TextArea(text, 5, 60, TextArea.SCROLLBARS_NONE);
   ta.setEditable(false); ta.setFocusable(false);
   tf = new TextField(s[0]);
   if (s[0].isEmpty()) tf.setVisible(false);
   else
   {  tf.requestFocusInWindow(); tf.selectAll();
      tf.addActionListener(this); addWindowListener(this);
   }
   add(ta, "North"); add(tf, "Center");
   Panel buttons = new Panel();
   b1 = new Button("Ok"); b1.setActionCommand("button1");
   b1.addActionListener(this); buttons.add(b1);
   b2 = new Button("Cancel"); b2.setActionCommand("button2");
   b2.addActionListener(this); buttons.add(b2);   
   add(buttons, "South");
   enableEvents(AWTEvent.KEY_EVENT_MASK);
}

public void processKeyEvent(KeyEvent e) //does not work, must use Alt+F4
{  int key = e.getKeyCode();
   if (key == KeyEvent.VK_ESCAPE)
   { s[0] = ""; setVisible(false); dispose(); }
   super.processKeyEvent(e);
}

public void actionPerformed(ActionEvent e)
{  if (e.getActionCommand() == "button2") s[0] = ""; else s[0] = tf.getText();
   setVisible(false); dispose();
}

public void windowClosing(WindowEvent e)
{ s[0] = ""; setVisible(false); dispose(); }

public void windowClosed(WindowEvent e) {}
public void windowDeactivated(WindowEvent e) {}
public void windowDeiconified(WindowEvent e) {}
public void windowIconified(WindowEvent e) {}
public void windowOpened(WindowEvent e) {}
public void windowActivated(WindowEvent e) {}
} //Stringdialog
